/*
 * @(#)hardware.c  1.1 1998/12/22 Carlos Hasan (chasan@dcc.uchile.cl)
 *
 * Hardware access routines for Sound Blaster 16 devices.
 *
 * Copyright (C) 1998 Carlos Hasan. All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include "hardware.h"

/*
 * ISA bus manager I/O access routines
 */
static struct isa_module_info *isa_module;

static status_t
init_isa_module(void)
{
    return get_module(B_ISA_MODULE_NAME, (module_info **) &isa_module);
}

static status_t
done_isa_module(void)
{
    return put_module(B_ISA_MODULE_NAME);
}

static uint8
read_io_8(int port)
{
    return isa_module->read_io_8(port);
}

static void
write_io_8(int port, uint8 value)
{
    isa_module->write_io_8(port, value);
}

static long
start_isa_dma(long channel, void *buf, long count, uchar mode, uchar emode)
{
    return isa_module->start_isa_dma(channel, buf, count, mode, emode);
}

static long
lock_isa_dma_channel(long channel)
{
    return isa_module->lock_isa_dma_channel(channel);
}

static long
unlock_isa_dma_channel(long channel)
{
    return isa_module->unlock_isa_dma_channel(channel);
}


/*
 * ISA Plug-and-Play config manager
 */
static status_t
find_sb16_isapnp_device(int *port, int *irq, int *dma8, int *dma16,
                        int *midi_port)
{
    config_manager_for_driver_module_info *config_module;
    uint64 cookie;
    struct device_info info;
    struct device_info *dev_info;
    struct isa_info *isa_info;
    struct device_configuration *dev_config;
    resource_descriptor resource[5];
    status_t err;
    int value;

    /* get config bus manager module */
    if ((err = get_module(B_CONFIG_MANAGER_FOR_DRIVER_MODULE_NAME,
			  (module_info **) &config_module)) < B_OK)
	return err;

    /* scan installed PnP ISA devices */
    cookie = 0;
    while (config_module->get_next_device_info(B_ISA_BUS, &cookie,
					       &info, sizeof(info)) == B_OK) {

	/* get ISA device information */
	dev_info = (struct device_info *) malloc(info.size);
	config_module->get_device_info_for(cookie, dev_info, info.size);
	isa_info = (struct isa_info *) ((char *) dev_info +
					dev_info->bus_dependent_info_offset);

	/* check PnP ISA logical device ID */
	if (!PNP_IS_SB16_DEVICE(isa_info->logical_device_id)) {
	    free(dev_info);
	    continue;
	}

	/* get PnP ISA device configuration structure */
	err = config_module->get_size_of_current_configuration_for(cookie);
	dev_config = (struct device_configuration *) malloc(err);
	config_module->get_current_configuration_for(cookie, dev_config, err);

	/* scan PnP ISA device configuration resources */
	if (config_module->get_nth_resource_descriptor_of_type(
	    dev_config, 0, B_IO_PORT_RESOURCE, &resource[0], sizeof(resource[0])) < B_OK ||
	    config_module->get_nth_resource_descriptor_of_type(
		dev_config, 1, B_IO_PORT_RESOURCE, &resource[4], sizeof(resource[4])) < B_OK ||
	    config_module->get_nth_resource_descriptor_of_type(
		dev_config, 0, B_IRQ_RESOURCE, &resource[1], sizeof(resource[1])) < B_OK ||
	    config_module->get_nth_resource_descriptor_of_type(
		dev_config, 0, B_DMA_RESOURCE, &resource[2], sizeof(resource[2])) < B_OK ||
	    config_module->get_nth_resource_descriptor_of_type(
		dev_config, 1, B_DMA_RESOURCE, &resource[3], sizeof(resource[3])) < B_OK) {
	    free(dev_config);
	    free(dev_info);
	    continue;
	}

	/* parse ISA device resource descriptors */
	*port = resource[0].d.r.minbase;
	*midi_port = resource[4].d.r.minbase;
	for (value = 0; value < 16; value++)
	    if (resource[1].d.m.mask & (1 << value))
		break;
	*irq = value;
	for (value = 0; value < 8; value++)
	    if (resource[2].d.m.mask & (1 << value))
		break;
	*dma8 = value;
	for (value = 0; value < 8; value++)
	    if (resource[3].d.m.mask & (1 << value))
		break;
	*dma16 = value;

	/* HACK: swap ViBRA 16X 8-bit DMA channels */
	if ((resource[2].d.m.cookie & 0x3) == 0x2 ||
	    (resource[3].d.m.cookie & 0x3) == 0x0) {
	    value = *dma8;
	    *dma8 = *dma16;
	    *dma16 = value;
	}

	/* release device structures */
	free(dev_config);
	free(dev_info);

	/* PnP ISA device found */
	put_module(B_CONFIG_MANAGER_FOR_DRIVER_MODULE_NAME);
	return B_OK;
    }

    /* shutdown config bus manager module */
    put_module(B_CONFIG_MANAGER_FOR_DRIVER_MODULE_NAME);
    return B_ERROR;
}

/*
 * Sound Blaster 16 DSP hardware configuration
 */
static int port, irq, dma8, dma16;
static struct {
    int bits;
    int channels;
    int latency;
    void (*callback)(void);
} play, rec;

static struct {
    int port;
    void (*callback)(void);
} mpu401;


/*
 * Sound Blaster 16 MPU-401 hardware interface routines
 */

status_t
hw_mpu401_write_command(int command)
{
    bigtime_t time;
  
    /* wait until the MPU-401 is ready */
    time = MPU401_IO_TIMEOUT;
    while (time > 0 && (read_io_8(mpu401.port + MPU401_STATUS) & B_MPU401_WRITE_BUSY)) {
	snooze(MPU401_IO_LATENCY);
	time -= MPU401_IO_LATENCY;
    }
    if (time <= 0)
	return B_DEV_TIMEOUT;
    
    /* write command to the MPU-401 port */
    write_io_8(mpu401.port + MPU401_COMMAND, command);
  
    /* wait until the MPU-401 is ready */
    time = MPU401_IO_TIMEOUT;
    while (time > 0 && (read_io_8(mpu401.port + MPU401_STATUS) & B_MPU401_READ_BUSY)) {
	snooze(MPU401_IO_LATENCY);
	time -= MPU401_IO_LATENCY;
    }
    if (time <= 0)
	return B_DEV_TIMEOUT;

    /* read acknowledge from the MPU-401 port */
    if (read_io_8(mpu401.port + MPU401_DATA) != 0xfe)
	return B_DEV_NOT_READY;
    
    return B_OK;
}

void
hw_mpu401_write_data(int data)
{
    write_io_8(mpu401.port + MPU401_DATA, data);
}
    
int
hw_mpu401_read_data(void)
{
    return read_io_8(mpu401.port + MPU401_DATA);
}

int
hw_mpu401_read_status(void)
{
    return read_io_8(mpu401.port + MPU401_STATUS);
}

int
hw_get_mpu401_port(void)
{
	return mpu401.port;
}

/*
 * Sound Blaster 16 DSP hardware interface routines
 */
static void
hw_codec_write_byte(uint8 value)
{
    int i;

    /* wait until the DSP is ready to receive data */
    for (i = 0; i < SB16_CODEC_IO_DELAY; i++) {
	if (!(read_io_8(port + SB16_CODEC_WRITE_STATUS) & 0x80))
	    break;
    }
    /* write a byte to the DSP data port */
    write_io_8(port + SB16_CODEC_WRITE_DATA, value);
}

static int
hw_codec_read_byte(void)
{
    int i;
    /* wait until the DSP has data available */
    for (i = 0; i < SB16_CODEC_IO_DELAY; i++) {
	if (read_io_8(port + SB16_CODEC_READ_STATUS) & 0x80)
	    break;
    }
    /* read a byte from the DSP data port */
    return read_io_8(port + SB16_CODEC_READ_DATA);
}

static void
hw_codec_reg_write(uint8 index, uint8 value)
{
    /* write a Mixer indirect register */
    write_io_8(port + SB16_MIXER_ADDRESS, index);
    write_io_8(port + SB16_MIXER_DATA, value);
}

static int
hw_codec_reg_read(uint8 index)
{
    /* read a Mixer indirect register */
    write_io_8(port + SB16_MIXER_ADDRESS, index);
    return read_io_8(port + SB16_MIXER_DATA);
}

static bool
hw_codec_reset(void)
{
    int times, delay;

    /* try to reset the DSP hardware */
    for (times = 0; times < 10; times++) {
	write_io_8(port + SB16_CODEC_RESET, 1);
	for (delay = 0; delay < SB16_CODEC_RESET_DELAY; delay++)
	    read_io_8(port + SB16_CODEC_RESET);
	write_io_8(port + SB16_CODEC_RESET, 0);
	if (hw_codec_read_byte() == 0xaa)
	    return true;
    }
    return false;
}

static int
hw_codec_read_version(void)
{
    int minor, major;

    /* query the DSP hardware version number */
    hw_codec_write_byte(SB16_CODEC_VERSION);
    major = hw_codec_read_byte();
    minor = hw_codec_read_byte();
    return (major << 8) + minor;
}

static void
hw_codec_read_irq_setup(int *irq)
{
    /* query the current IRQ line resource */
    int mask = hw_codec_reg_read(SB16_IRQ_SETUP);
    *irq = 5;
    if (mask & 0x01)
	*irq = 2;
    if (mask & 0x02)
	*irq = 5;
    if (mask & 0x04)
	*irq = 7;
    if (mask & 0x08)
	*irq = 10;
}

static void
hw_codec_read_dma_setup(int *dma8, int *dma16)
{
    /* query the current DMA channel resources */
    int mask = hw_codec_reg_read(SB16_DMA_SETUP);

    *dma8 = 1;
    if (mask & 0x01)
	*dma8 = 0;
    if (mask & 0x02)
	*dma8 = 1;
    if (mask & 0x08)
	*dma8 = 3;

    *dma16 = *dma8;
    if (mask & 0x20)
	*dma16 = 5;
    if (mask & 0x40)
	*dma16 = 6;
    if (mask & 0x80)
	*dma16 = 7;
}

static void
hw_codec_write_irq_setup(int irq)
{
    /* change programmable IRQ line resource */
    int mask = 0x02;
    if (irq == 2)
	mask = 0x01;
    if (irq == 5)
	mask = 0x02;
    if (irq == 7)
	mask = 0x04;
    if (irq == 10)
	mask = 0x08;
    hw_codec_reg_write(SB16_IRQ_SETUP, mask);
}

static void
hw_codec_write_dma_setup(int dma8, int dma16)
{
    /* change programmable DMA channel resources */
    hw_codec_reg_write(SB16_DMA_SETUP, (1 << dma8) | (1 << dma16));
}

static int32
hw_codec_inth(void *data)
{
    /* read the IRQ interrupt status register */
    int status = hw_codec_reg_read(SB16_IRQ_STATUS);

  /* check if this hardware raised this interrupt */
    if (status & 0x03) {
	/* acknowledge DMA memory transfers */
	if (status & 0x01)
	    read_io_8(port + SB16_CODEC_ACK_8_BIT);
	if (status & 0x02)
	    read_io_8(port + SB16_CODEC_ACK_16_BIT);

	/* acknowledge PIC interrupt signal */
	if (irq >= 8)
	    write_io_8(0xa0, 0x20);
	write_io_8(0x20, 0x20);

	/* call the user interrupt routine */
	if (((play.bits >> 3) & status) != 0 && play.callback != 0)
	    play.callback();
	if (((rec.bits >> 3) & status) != 0 && rec.callback != 0)
	    rec.callback();
	if ((status & 0x04) != 0 && mpu401.callback != 0)
	    mpu401.callback();
      
	return B_INVOKE_SCHEDULER;
    }

    /* we didn't raise this interrupt */
    return B_UNHANDLED_INTERRUPT;
}

static bool
hw_codec_detect(void)
{
    int fd, n;
    char *s, buffer[32];

    /* setup default settings */
    port = 0x220;
    irq = 5;
    dma8 = 1;
    dma16 = 5;
    mpu401.port = 0x330;

    /* reset playback and capture settings */
    play.bits = play.channels = play.latency = 0;
    rec.bits = rec.channels = rec.latency = 0;

    /* reset playback, capture and midi callbacks */
    play.callback = rec.callback = mpu401.callback = 0;

    /* read user hardware settings */
    if ((fd = open("/boot/home/config/settings/sb16_settings", O_RDONLY)) > 0) {
	n = read(fd, buffer, sizeof(buffer));
	close(fd);
	if (n > 0) {
	    buffer[n] = 0;
	    if ((s = strchr(buffer, 'A')) != NULL)
		port = strtol(s + 1, NULL, 16);
	    if ((s = strchr(buffer, 'I')) != NULL)
		irq = strtol(s + 1, NULL, 10);
	    if ((s = strchr(buffer, 'D')) != NULL)
		dma8 = strtol(s + 1, NULL, 10);
	    if ((s = strchr(buffer, 'H')) != NULL)
		dma16 = strtol(s + 1, NULL, 10);
	    if ((s = strchr(buffer, 'P')) != NULL)
		mpu401.port = strtol(s + 1, NULL, 16);
	    if (hw_codec_reset() && hw_codec_read_version() >= 0x400) {
		hw_codec_write_irq_setup(irq);
		hw_codec_write_dma_setup(dma8, dma16);
		return true;
	    }
	    return false;
	}
    }

    /* search Creative SB16 ISA PnP audio device */
    if (find_sb16_isapnp_device(&port, &irq, &dma8, &dma16, &mpu401.port) == B_OK)
	return (hw_codec_reset() && hw_codec_read_version() >= 0x400);

    /* search the hardware on a range of ports */
    for (port = 0x220; port <= 0x280; port += 0x20) {
	/* there is a DSP hardware version 4.0 or better? */
	if (hw_codec_reset() && hw_codec_read_version() >= 0x400) {
	    /* query and change the IRQ line resource */
	    hw_codec_read_irq_setup(&irq);
	    hw_codec_write_irq_setup(irq);

	    /* query and change the DMA channel resources */
	    hw_codec_read_dma_setup(&dma8, &dma16);
	    hw_codec_write_dma_setup(dma8, dma16);

	    /* detect MPU-401 base I/O port address */
	    for (mpu401.port = 0x330; mpu401.port >= 0x300; mpu401.port -= 0x30) {
		if (hw_mpu401_write_command(MPU401_CMD_RESET) == B_OK)
		    return true;
	    }
	}
    }
    return false;
}

/*
 * Sound Blaster 16 high-level hardware access routines
 */
status_t
hw_detect(void)
{
    /* initialize ISA bus manager */
    if (init_isa_module() < B_OK)
	return B_ERROR;

  /* search and configure the hardware */
    if (hw_codec_detect()) {
	/* clear pending interrupts */
	read_io_8(port + SB16_CODEC_ACK_8_BIT);
	read_io_8(port + SB16_CODEC_ACK_16_BIT);
	done_isa_module();
	return B_NO_ERROR;
    }

    /* hardware not found */
    done_isa_module();
    return B_ERROR;
}

status_t
hw_init(void)
{
    /* initialize ISA bus manager */
    if (init_isa_module() < B_OK)
	return B_ERROR;

  /* reset the audio hardware */
    if (hw_codec_detect()) {
	/* reserve ISA DMA channel resources */
	if (lock_isa_dma_channel(dma8) == B_NO_ERROR &&
	    (dma8 != dma16 || lock_isa_dma_channel(dma16) == B_NO_ERROR)) {
	    /* install the I/O interrupt handler */
	    install_io_interrupt_handler(irq, hw_codec_inth, NULL, 0);
	    return B_NO_ERROR;
	}
    }
    return B_ERROR;
}

void
hw_done(void)
{
    /* remove the I/O interrupt handler */
    remove_io_interrupt_handler(irq, hw_codec_inth, NULL);

  /* release ISA DMA channel resources */
    if (dma8 != dma16)
	unlock_isa_dma_channel(dma16);
    unlock_isa_dma_channel(dma8);

    /* reset audio hardware */
    hw_codec_reset();

    /* shutdown ISA bus manager */
    done_isa_module();
}

void
hw_set_playback_callback(void (*callback)(void))
{
    /* change playback interrupt routine */
    play.callback = callback;
}

void
hw_set_capture_callback(void (*callback)(void))
{
    /* change capture interrupt routine */
    rec.callback = callback;
}

void
hw_set_mpu401_callback(void (*callback)(void))
{
    /* change midi interrupt routine */
    mpu401.callback = callback;
}

void
hw_clear_playback_interrupts(void)
{
    /* clear pending interrupts */
    if (play.bits == 8)
	read_io_8(port + SB16_CODEC_ACK_8_BIT);
    else
	read_io_8(port + SB16_CODEC_ACK_16_BIT);
}

void
hw_clear_capture_interrupts(void)
{
    /* clear pending interrupts */
    if (rec.bits == 8)
	read_io_8(port + SB16_CODEC_ACK_8_BIT);
    else
	read_io_8(port + SB16_CODEC_ACK_16_BIT);
}

void
hw_set_playback_rate(int sample_rate)
{
    /* set the playback sampling rate (5000 to 44100 Hz) */
    hw_codec_write_byte(SB16_PLAYBACK_RATE);
    hw_codec_write_byte(sample_rate >> 8);
    hw_codec_write_byte(sample_rate & 255);
}

void
hw_set_capture_rate(int sample_rate)
{
    /* set the capture sampling rate (5000 to 44100 Hz) */
    hw_codec_write_byte(SB16_CAPTURE_RATE);
    hw_codec_write_byte(sample_rate >> 8);
    hw_codec_write_byte(sample_rate & 255);
}

status_t
hw_alloc_buffer(hw_dma_buffer *buffer, int size)
{
    status_t err;
    physical_entry phys;

    /* initialize buffer structure */
    buffer->area = 0;
    buffer->address = NULL;
    buffer->size = B_PAGE_SIZE * ((size + B_PAGE_SIZE - 1) / B_PAGE_SIZE);

    /* create I/O transfer area in low memory */
    if ((err = buffer->area = create_area("sb16 buffer",
					  (void **) &buffer->address, B_ANY_KERNEL_ADDRESS, buffer->size,
					  B_LOMEM, B_READ_AREA | B_WRITE_AREA)) < B_OK)
	return err;

    /* adjust I/O transfer area address */
    get_memory_map(buffer->address, buffer->size, &phys, 1);
    phys.size = 0x10000L - ((uint32) phys.address & 0xffff);
    if (phys.size < buffer->size) {
	if (phys.size > buffer->size / 2)
	    buffer->size = phys.size;
	else {
	    buffer->address += phys.size;
	    buffer->size -= phys.size;
	}
    }

    /* lock I/O transfer area in memory */
    lock_memory(buffer->address, buffer->size, B_DMA_IO | B_READ_DEVICE);

    return B_OK;
}

void
hw_free_buffer(hw_dma_buffer *buffer)
{
    /* unlock I/O transfer area from memory */
    unlock_memory(buffer->address, buffer->size, B_DMA_IO | B_READ_DEVICE);

  /* release I/O transfer area buffer */
    delete_area(buffer->area);
}

void
hw_set_playback_buffer(hw_dma_buffer *buffer, int bits, int channels, int latency)
{
    physical_entry phys;

    /* remember playback settings */
    play.bits = bits;
    play.channels = channels;
    play.latency = latency;

    /* get physical address of the buffer */
    get_memory_map(buffer->address, buffer->size, &phys, 1);

    /* start auto-init DMA memory transfer */
    if (play.bits == 8)
	start_isa_dma(dma8, phys.address, phys.size, 0x58, 0);
    else
	start_isa_dma(dma16, phys.address, phys.size >> (dma16 >> 2), 0x58, 0);
}

void
hw_set_capture_buffer(hw_dma_buffer *buffer, int bits, int channels, int latency)
{
    physical_entry phys;

    /* remember capture settings */
    rec.bits = bits;
    rec.channels = channels;
    rec.latency = latency;

    /* get physical address of the buffer */
    get_memory_map(buffer->address, buffer->size, &phys, 1);

    /* start auto-init DMA memory transfer */
    if (rec.bits == 8)
	start_isa_dma(dma8, phys.address, phys.size, 0x54, 0);
    else
	start_isa_dma(dma16, phys.address, phys.size >> (dma16 >> 2), 0x54, 0);
}

void
hw_start_playback(void)
{
    /* program auto-init DMA sound I/O playback */
    if (play.bits == 8)
	hw_codec_write_byte(SB16_PLAYBACK_8_BIT);
    else
	hw_codec_write_byte(SB16_PLAYBACK_16_BIT);

  /* select auto-init DMA sound I/O format */
    hw_codec_write_byte(
	(play.bits == 8 ? B_SB16_FORMAT_UNSIGNED : B_SB16_FORMAT_SIGNED) +
	(play.channels == 1 ? B_SB16_FORMAT_MONO : B_SB16_FORMAT_STEREO));

    /* set auto-init DMA sound I/O latency */
    if (play.bits == 8) {
	hw_codec_write_byte((play.latency - 1) & 255);
	hw_codec_write_byte((play.latency - 1) >> 8);
    }
    else {
	hw_codec_write_byte(((play.latency >> 1) - 1) & 255);
	hw_codec_write_byte(((play.latency >> 1) - 1) >> 8);
    }
}

void
hw_start_capture(void)
{
    /* program auto-init DMA sound I/O capture */
    if (rec.bits == 8)
	hw_codec_write_byte(SB16_CAPTURE_8_BIT);
    else
	hw_codec_write_byte(SB16_CAPTURE_16_BIT);

  /* select auto-init DMA sound I/O format */
    hw_codec_write_byte(
	(rec.bits == 8 ? B_SB16_FORMAT_UNSIGNED : B_SB16_FORMAT_SIGNED) +
	(rec.channels == 1 ? B_SB16_FORMAT_MONO : B_SB16_FORMAT_STEREO));

    /* set auto-init DMA sound I/O latency */
    if (rec.bits == 8) {
	hw_codec_write_byte((rec.latency - 1) & 255);
	hw_codec_write_byte((rec.latency - 1) >> 8);
    }
    else {
	hw_codec_write_byte(((rec.latency >> 1) - 1) & 255);
	hw_codec_write_byte(((rec.latency >> 1) - 1) >> 8);
    }
}

void
hw_stop_playback(void)
{
    /* exit 8 or 16 bit autoinit DMA mode sound I/O */
    if (play.bits == 8)
	hw_codec_write_byte(SB16_DMA_EXIT_8_BIT);
    else
	hw_codec_write_byte(SB16_DMA_EXIT_16_BIT);
}

void
hw_stop_capture(void)
{
    /* exit 8 or 16 bit autoinit DMA mode sound I/O */
    if (rec.bits == 8)
	hw_codec_write_byte(SB16_DMA_EXIT_8_BIT);
    else
	hw_codec_write_byte(SB16_DMA_EXIT_16_BIT);
}

void
hw_pause_playback(void)
{
    /* pause 8 or 16 bit DMA mode sound I/O */
    if (play.bits == 8)
	hw_codec_write_byte(SB16_DMA_DISABLE_8_BIT);
    else
	hw_codec_write_byte(SB16_DMA_DISABLE_16_BIT);
}

void
hw_pause_capture(void)
{
    /* pause 8 or 16 bit DMA mode sound I/O */
    if (rec.bits == 8)
	hw_codec_write_byte(SB16_DMA_DISABLE_8_BIT);
    else
	hw_codec_write_byte(SB16_DMA_DISABLE_16_BIT);
}

void
hw_resume_playback(void)
{
    /* continue 8 or 16 bit DMA mode sound I/O */
    if (play.bits == 8)
	hw_codec_write_byte(SB16_DMA_ENABLE_8_BIT);
    else
	hw_codec_write_byte(SB16_DMA_ENABLE_16_BIT);
}

void
hw_resume_capture(void)
{
    /* continue 8 or 16 bit DMA mode sound I/O */
    if (rec.bits == 8)
	hw_codec_write_byte(SB16_DMA_ENABLE_8_BIT);
    else
	hw_codec_write_byte(SB16_DMA_ENABLE_16_BIT);
}

void
hw_set_mixer_value(int selector, int value)
{
    hw_codec_reg_write(selector, value);
}

int
hw_get_mixer_value(int selector)
{
    return hw_codec_reg_read(selector);
}

status_t
hw_reset_stereo_enhancement(void)
{
    /* check if stereo enhancement is available */
    hw_set_mixer_value(SB16_STEREO_ENHANCEMENT, B_SB16_SE_DETECT_ENABLE);
    if (hw_get_mixer_value(SB16_STEREO_ENHANCEMENT) & B_SB16_SE_DETECT_STATUS)
	return B_ERROR;

  /* reset stereo enhancement output */
    hw_set_mixer_value(SB16_STEREO_ENHANCEMENT, 0);
    return B_NO_ERROR;
}

